@@ -1,7 +1,16 @@
 /*
- * Copyright (C) 2001-2003 Sistina Software (UK) Limited.
+ * Copyright (C) 2001-2004 Sistina Software, Inc. All rights reserved.
+ * Copyright (C) 2004 Red Hat, Inc. All rights reserved.
  *
- * This file is released under the GPL.
+ * This file is part of the device-mapper userspace tools.
+ *
+ * This copyrighted material is made available to anyone wishing to use,
+ * modify, copy, or redistribute it subject to the terms and conditions
+ * of the GNU General Public License v.2.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software Foundation,
+ * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
 #include "libdevmapper.h"
@@ -63,7 +72,7 @@ static char *_uuid;
  */
 static int _parse_file(struct dm_task *dmt, const char *file)
 {
-	char buffer[LINE_SIZE], *ttype, *ptr, *comment;
+	char buffer[LINE_SIZE], ttype[LINE_SIZE], *ptr, *comment;
 	FILE *fp;
 	unsigned long long start, size;
 	int r = 0, n, line = 0;
@@ -94,8 +103,8 @@ static int _parse_file(struct dm_task *dmt, const char *file)
 		if (!*ptr || *ptr == '#')
 			continue;
 
-		if (sscanf(ptr, "%llu %llu %as %n",
-			   &start, &size, &ttype, &n) < 3) {
+		if (sscanf(ptr, "%llu %llu %s %n",
+			   &start, &size, ttype, &n) < 3) {
 			err("%s:%d Invalid format", file, line);
 			goto out;
 		}
@@ -106,8 +115,6 @@ static int _parse_file(struct dm_task *dmt, const char *file)
 
 		if (!dm_task_add_target(dmt, start, size, ttype, ptr))
 			goto out;
-
-		free(ttype);
 	}
 	r = 1;
 
@@ -327,40 +334,6 @@ static int _wait(int argc, char **argv, void *data)
 		       (argc == 3) ? atoi(argv[2]) : 0, 1);
 }
 
-static int _process_mapper_dir(int argc, char **argv,
-			       int (*fn) (int argc, char **argv, void *data))
-{
-	struct dirent *dirent;
-	struct dm_names *names;
-	DIR *d;
-	const char *dir;
-	int r = 1;
-
-	dir = dm_dir();
-	if (!(d = opendir(dir))) {
-		fprintf(stderr, "opendir %s: %s", dir, strerror(errno));
-		return 0;
-	}
-
-	while ((dirent = readdir(d))) {
-		if (!strcmp(dirent->d_name, ".") ||
-		    !strcmp(dirent->d_name, "..") ||
-		    !strcmp(dirent->d_name, "control"))
-			continue;
-		/* Set up names->name for _info */
-		names = (void *) dirent->d_name -
-		    ((void *) &names->name - (void *) &names->dev);
-		if (!fn(argc, argv, names))
-			r = 0;
-	}
-
-	if (closedir(d)) {
-		fprintf(stderr, "closedir %s: %s", dir, strerror(errno));
-	}
-
-	return r;
-}
-
 static int _process_all(int argc, char **argv,
 			int (*fn) (int argc, char **argv, void *data))
 {
@@ -370,9 +343,6 @@ static int _process_all(int argc, char **argv,
 
 	struct dm_task *dmt;
 
-	if (!strcmp(argv[0], "mknodes"))
-		r = _process_mapper_dir(argc, argv, fn);
-
 	if (!(dmt = dm_task_create(DM_DEVICE_LIST)))
 		return 0;
 
@@ -497,14 +467,34 @@ static int _targets(int argc, char **argv, void *data)
 
 }
 
+static int _mknodes(int argc, char **argv, void *data)
+{
+	struct dm_task *dmt;
+	int r = 0;
+
+	if (!(dmt = dm_task_create(DM_DEVICE_MKNODES)))
+		return 0;
+
+	if (argc == 2 && !dm_task_set_name(dmt, argv[1]))
+		goto out;
+
+	if (!dm_task_run(dmt))
+		goto out;
+
+	r = 1;
+
+      out:
+	dm_task_destroy(dmt);
+	return r;
+}
+
 static int _info(int argc, char **argv, void *data)
 {
 	int r = 0;
 
 	struct dm_task *dmt;
 	struct dm_names *names = (struct dm_names *) data;
-	char *name;
-	int taskno;
+	char *name = NULL;
 
 	if (argc == 1 && !data)
 		return _process_all(argc, argv, _info);
@@ -514,22 +504,16 @@ static int _info(int argc, char **argv, void *data)
 	else
 		name = argv[1];
 
-	if (!strcmp(argv[0], "mknodes"))
-		taskno = DM_DEVICE_MKNODES;
-	else
-		taskno = DM_DEVICE_INFO;
-
-	if (!(dmt = dm_task_create(taskno)))
+	if (!(dmt = dm_task_create(DM_DEVICE_INFO)))
 		return 0;
 
-	if (!dm_task_set_name(dmt, name))
+	if (name && !dm_task_set_name(dmt, name))
 		goto out;
 
 	if (!dm_task_run(dmt))
 		goto out;
 
-	if (taskno == DM_DEVICE_INFO)
-		_display_info(dmt);
+	_display_info(dmt);
 
 	r = 1;
 
@@ -642,10 +626,10 @@ static struct command _commands[] = {
 	{"ls", "", 0, 0, _ls},
 	{"info", "[<dev_name>]", 0, 1, _info},
 	{"deps", "[<dev_name>]", 0, 1, _deps},
-	{"mknodes", "[<dev_name>]", 0, 1, _info},
 	{"status", "[<dev_name>]", 0, 1, _status},
 	{"table", "[<dev_name>]", 0, 1, _status},
 	{"wait", "<dev_name> [<event_nr>]", 1, 2, _wait},
+	{"mknodes", "[<dev_name>]", 0, 1, _mknodes},
 	{"targets", "", 0, 0, _targets},
 	{"version", "", 0, 0, _version},
 	{NULL, NULL, 0, 0, NULL}
