@@ -5,6 +5,7 @@
  */
 
 #include "dm.h"
+#include "kcopyd.h"
 
 #include <linux/init.h>
 #include <linux/module.h>
@@ -12,22 +13,21 @@
 #include <linux/blkpg.h>
 #include <linux/mempool.h>
 #include <linux/slab.h>
+#include <linux/major.h>
 #include <linux/kdev_t.h>
 #include <linux/lvm.h>
 
 #include <asm/uaccess.h>
 
 static const char *_name = DM_NAME;
-#define MAX_DEVICES (1 << MINORBITS)
-#define SECTOR_SHIFT 9
 #define DEFAULT_READ_AHEAD 64
 
-static int major = 0;
-static int _major = 0;
-
 struct dm_io {
 	struct mapped_device *md;
 
+	struct dm_target *ti;
+	int rw;
+	union map_info map_context;
 	void (*end_io) (struct buffer_head * bh, int uptodate);
 	void *context;
 };
@@ -62,25 +62,277 @@ struct mapped_device {
 	 * The current mapping.
 	 */
 	struct dm_table *map;
+
+	/*
+	 * io objects are allocated from here.
+	 */
+	mempool_t *io_pool;
+
+	/*
+	 * Event handling.
+	 */
+	uint32_t event_nr;
+	wait_queue_head_t eventq;
 };
 
 #define MIN_IOS 256
 static kmem_cache_t *_io_cache;
-static mempool_t *_io_pool;
-
-/* block device arrays */
-static int _block_size[MAX_DEVICES];
-static int _blksize_size[MAX_DEVICES];
-static int _hardsect_size[MAX_DEVICES];
 
 static struct mapped_device *get_kdev(kdev_t dev);
 static int dm_request(request_queue_t *q, int rw, struct buffer_head *bh);
 static int dm_user_bmap(struct inode *inode, struct lv_bmap *lvb);
 
+/*-----------------------------------------------------------------
+ * In order to avoid the 256 minor number limit we are going to
+ * register more major numbers as neccessary.
+ *---------------------------------------------------------------*/
+#define MAX_MINORS (1 << MINORBITS)
+
+struct major_details {
+	unsigned int major;
+
+	int transient;
+	struct list_head transient_list;
+
+	unsigned int first_free_minor;
+	int nr_free_minors;
+
+	struct mapped_device *mds[MAX_MINORS];
+	int blk_size[MAX_MINORS];
+	int blksize_size[MAX_MINORS];
+	int hardsect_size[MAX_MINORS];
+};
+
+static struct rw_semaphore _dev_lock;
+static struct major_details *_majors[MAX_BLKDEV];
+
+/*
+ * This holds a list of majors that non-specified device numbers
+ * may be allocated from.  Only majors with free minors appear on
+ * this list.
+ */
+static LIST_HEAD(_transients_free);
+
+static int __alloc_major(unsigned int major, struct major_details **result)
+{
+	int r;
+	unsigned int transient = !major;
+	struct major_details *maj;
+
+	/* Major already allocated? */
+	if (major && _majors[major])
+		return 0;
+
+	maj = kmalloc(sizeof(*maj), GFP_KERNEL);
+	if (!maj)
+		return -ENOMEM;
+
+	memset(maj, 0, sizeof(*maj));
+	INIT_LIST_HEAD(&maj->transient_list);
+
+	maj->nr_free_minors = MAX_MINORS;
+
+	r = register_blkdev(major, _name, &dm_blk_dops);
+	if (r < 0) {
+		DMERR("register_blkdev failed for %d", major);
+		kfree(maj);
+		return r;
+	}
+	if (r > 0)
+		major = r;
+
+	maj->major = major;
+
+	if (transient) {
+		maj->transient = transient;
+		list_add_tail(&maj->transient_list, &_transients_free);
+	}
+
+	_majors[major] = maj;
+
+	blk_size[major] = maj->blk_size;
+	blksize_size[major] = maj->blksize_size;
+	hardsect_size[major] = maj->hardsect_size;
+	read_ahead[major] = DEFAULT_READ_AHEAD;
+
+	blk_queue_make_request(BLK_DEFAULT_QUEUE(major), dm_request);
+
+	*result = maj;
+	return 0;
+}
+
+static void __free_major(struct major_details *maj)
+{
+	unsigned int major = maj->major;
+
+	list_del(&maj->transient_list);
+
+	read_ahead[major] = 0;
+	blk_size[major] = NULL;
+	blksize_size[major] = NULL;
+	hardsect_size[major] = NULL;
+
+	_majors[major] = NULL;
+	kfree(maj);
+
+	if (unregister_blkdev(major, _name) < 0)
+		DMERR("devfs_unregister_blkdev failed");
+}
+
+static void free_all_majors(void)
+{
+	unsigned int major = ARRAY_SIZE(_majors);
+
+	down_write(&_dev_lock);
+
+	while (major--)
+		if (_majors[major])
+			__free_major(_majors[major]);
+
+	up_write(&_dev_lock);
+}
+
+static void free_dev(kdev_t dev)
+{
+	unsigned int major = major(dev);
+	unsigned int minor = minor(dev);
+	struct major_details *maj;
+
+	down_write(&_dev_lock);
+
+	maj = _majors[major];
+	if (!maj)
+		goto out;
+
+	maj->mds[minor] = NULL;
+	maj->nr_free_minors++;
+
+	if (maj->nr_free_minors == MAX_MINORS) {
+		__free_major(maj);
+		goto out;
+	}
+
+	if (!maj->transient)
+		goto out;
+
+	if (maj->nr_free_minors == 1)
+		list_add_tail(&maj->transient_list, &_transients_free);
+
+	if (minor < maj->first_free_minor)
+		maj->first_free_minor = minor;
+
+      out:
+	up_write(&_dev_lock);
+}
+
+static void __alloc_minor(struct major_details *maj, unsigned int minor,
+			  struct mapped_device *md)
+{
+	maj->mds[minor] = md;
+	md->dev = mk_kdev(maj->major, minor);
+	maj->nr_free_minors--;
+
+	if (maj->transient && !maj->nr_free_minors)
+		list_del_init(&maj->transient_list);
+}
+
+/*
+ * See if requested kdev_t is available.
+ */
+static int specific_dev(kdev_t dev, struct mapped_device *md)
+{
+	int r = 0;
+	unsigned int major = major(dev);
+	unsigned int minor = minor(dev);
+	struct major_details *maj;
+
+	if (!major || (major > MAX_BLKDEV) || (minor >= MAX_MINORS)) {
+		DMWARN("device number requested out of range (%d, %d)",
+		       major, minor);
+		return -EINVAL;
+	}
+
+	down_write(&_dev_lock);
+	maj = _majors[major];
+
+	/* Register requested major? */
+	if (!maj) {
+		r = __alloc_major(major, &maj);
+		if (r)
+			goto out;
+
+		major = maj->major;
+	}
+
+	if (maj->mds[minor]) {
+		r = -EBUSY;
+		goto out;
+	}
+
+	__alloc_minor(maj, minor, md);
+
+      out:
+	up_write(&_dev_lock);
+
+	return r;
+}
+
+/*
+ * Find first unused device number, requesting a new major number if required.
+ */
+static int first_free_dev(struct mapped_device *md)
+{
+	int r = 0;
+	struct major_details *maj;
+
+	down_write(&_dev_lock);
+
+	if (list_empty(&_transients_free)) {
+		r = __alloc_major(0, &maj);
+		if (r)
+			goto out;
+	} else
+		maj = list_entry(_transients_free.next, struct major_details,
+				 transient_list);
+
+	while (maj->mds[maj->first_free_minor++])
+		;
+
+	__alloc_minor(maj, maj->first_free_minor - 1, md);
+
+      out:
+	up_write(&_dev_lock);
+
+	return r;
+}
+
+static struct mapped_device *get_kdev(kdev_t dev)
+{
+	struct mapped_device *md;
+	struct major_details *maj;
+
+	down_read(&_dev_lock);
+	maj = _majors[major(dev)];
+	if (!maj) {
+		md = NULL;
+		goto out;
+	}
+	md = maj->mds[minor(dev)];
+	if (md)
+		dm_get(md);
+      out:
+	up_read(&_dev_lock);
+
+	return md;
+}
+
+/*-----------------------------------------------------------------
+ * init/exit code
+ *---------------------------------------------------------------*/
 
 static __init int local_init(void)
 {
-	int r;
+	init_rwsem(&_dev_lock);
 
 	/* allocate a slab for the dm_ios */
 	_io_cache = kmem_cache_create("dm io",
@@ -89,49 +341,13 @@ static __init int local_init(void)
 	if (!_io_cache)
 		return -ENOMEM;
 
-	_io_pool = mempool_create(MIN_IOS, mempool_alloc_slab,
-				  mempool_free_slab, _io_cache);
-	if (!_io_pool) {
-		kmem_cache_destroy(_io_cache);
-		return -ENOMEM;
-	}
-
-	_major = major;
-	r = register_blkdev(_major, _name, &dm_blk_dops);
-	if (r < 0) {
-		DMERR("register_blkdev failed");
-		mempool_destroy(_io_pool);
-		kmem_cache_destroy(_io_cache);
-		return r;
-	}
-
-	if (!_major)
-		_major = r;
-
-	/* set up the arrays */
-	read_ahead[_major] = DEFAULT_READ_AHEAD;
-	blk_size[_major] = _block_size;
-	blksize_size[_major] = _blksize_size;
-	hardsect_size[_major] = _hardsect_size;
-
-	blk_queue_make_request(BLK_DEFAULT_QUEUE(_major), dm_request);
-
 	return 0;
 }
 
 static void local_exit(void)
 {
-	mempool_destroy(_io_pool);
 	kmem_cache_destroy(_io_cache);
-
-	if (unregister_blkdev(_major, _name) < 0)
-		DMERR("devfs_unregister_blkdev failed");
-
-	read_ahead[_major] = 0;
-	blk_size[_major] = NULL;
-	blksize_size[_major] = NULL;
-	hardsect_size[_major] = NULL;
-	_major = 0;
+	free_all_majors();
 
 	DMINFO("cleaned up");
 }
@@ -148,6 +364,7 @@ static struct {
 } _inits[] = {
 #define xx(n) {n ## _init, n ## _exit},
 	xx(local)
+	xx(kcopyd)
 	xx(dm_target)
 	xx(dm_linear)
 	xx(dm_stripe)
@@ -209,14 +426,14 @@ static int dm_blk_close(struct inode *inode, struct file *file)
 	return 0;
 }
 
-static inline struct dm_io *alloc_io(void)
+static inline struct dm_io *alloc_io(struct mapped_device *md)
 {
-	return mempool_alloc(_io_pool, GFP_NOIO);
+	return mempool_alloc(md->io_pool, GFP_NOIO);
 }
 
-static inline void free_io(struct dm_io *io)
+static inline void free_io(struct mapped_device *md, struct dm_io *io)
 {
-	mempool_free(io, _io_pool);
+	mempool_free(io, md->io_pool);
 }
 
 static inline struct deferred_io *alloc_deferred(void)
@@ -229,19 +446,18 @@ static inline void free_deferred(struct deferred_io *di)
 	kfree(di);
 }
 
-/* In 512-byte units */
-#define VOLUME_SIZE(minor) (_block_size[(minor)] << 1)
+static inline sector_t volume_size(kdev_t dev)
+{
+	return blk_size[major(dev)][minor(dev)] << 1;
+}
 
 /* FIXME: check this */
 static int dm_blk_ioctl(struct inode *inode, struct file *file,
-			uint command, unsigned long a)
+			unsigned int command, unsigned long a)
 {
-	int minor = MINOR(inode->i_rdev);
+	kdev_t dev = inode->i_rdev;
 	long size;
 
-	if (minor >= MAX_DEVICES)
-		return -ENXIO;
-
 	switch (command) {
 	case BLKROSET:
 	case BLKROGET:
@@ -255,17 +471,17 @@ static int dm_blk_ioctl(struct inode *inode, struct file *file,
 	case BLKELVSET:
 	case BLKBSZGET:
 	case BLKBSZSET:
-		return blk_ioctl(inode->i_rdev, command, a);
+		return blk_ioctl(dev, command, a);
 		break;
 
 	case BLKGETSIZE:
-		size = VOLUME_SIZE(minor);
+		size = volume_size(dev);
 		if (copy_to_user((void *) a, &size, sizeof(long)))
 			return -EFAULT;
 		break;
 
 	case BLKGETSIZE64:
-		size = VOLUME_SIZE(minor);
+		size = volume_size(dev);
 		if (put_user((u64) ((u64) size) << 9, (u64 *) a))
 			return -EFAULT;
 		break;
@@ -297,7 +513,7 @@ static int queue_io(struct mapped_device *md, struct buffer_head *bh, int rw)
 
 	down_write(&md->lock);
 
-	if (!test_bit(DMF_SUSPENDED, &md->flags)) {
+	if (!test_bit(DMF_BLOCK_IO, &md->flags)) {
 		up_write(&md->lock);
 		free_deferred(di);
 		return 1;
@@ -318,7 +534,20 @@ static int queue_io(struct mapped_device *md, struct buffer_head *bh, int rw)
  */
 static void dec_pending(struct buffer_head *bh, int uptodate)
 {
+	int r;
 	struct dm_io *io = bh->b_private;
+	dm_endio_fn endio = io->ti->type->end_io;
+
+	if (endio) {
+		r = endio(io->ti, bh, io->rw, uptodate ? 0 : -EIO,
+			  &io->map_context);
+		if (r < 0)
+			uptodate = 0;
+
+		else if (r > 0)
+			/* the target wants another shot at the io */
+			return;
+	}
 
 	if (atomic_dec_and_test(&io->md->pending))
 		/* nudge anyone waiting on suspend queue */
@@ -326,7 +555,7 @@ static void dec_pending(struct buffer_head *bh, int uptodate)
 
 	bh->b_end_io = io->end_io;
 	bh->b_private = io->context;
-	free_io(io);
+	free_io(io->md, io);
 
 	bh->b_end_io(bh, uptodate);
 }
@@ -334,38 +563,29 @@ static void dec_pending(struct buffer_head *bh, int uptodate)
 /*
  * Do the bh mapping for a given leaf
  */
-static inline int __map_buffer(struct mapped_device *md,
-			       int rw, struct buffer_head *bh)
+static inline int __map_buffer(struct mapped_device *md, int rw,
+			       struct buffer_head *bh, struct dm_io *io)
 {
-	int r;
-	struct dm_io *io;
 	struct dm_target *ti;
 
-	ti = dm_table_find_target(md->map, bh->b_rsector);
-	if (!ti)
+	if (!md->map)
 		return -EINVAL;
 
-	io = alloc_io();
-	if (!io)
-		return -ENOMEM;
+	ti = dm_table_find_target(md->map, bh->b_rsector);
+	if (!ti->type)
+		return -EINVAL;
 
+	/* hook the end io request fn */
+	atomic_inc(&md->pending);
 	io->md = md;
+	io->ti = ti;
+	io->rw = rw;
 	io->end_io = bh->b_end_io;
 	io->context = bh->b_private;
+	bh->b_end_io = dec_pending;
+	bh->b_private = io;
 
-	r = ti->type->map(ti, bh, rw);
-
-	if (r > 0) {
-		/* hook the end io request fn */
-		atomic_inc(&md->pending);
-		bh->b_end_io = dec_pending;
-		bh->b_private = io;
-
-	} else
-		/* we don't need to hook */
-		free_io(io);
-
-	return r;
+	return ti->type->map(ti, bh, rw, &io->map_context);
 }
 
 /*
@@ -409,6 +629,7 @@ static inline int __deferring(struct mapped_device *md, int rw,
 static int dm_request(request_queue_t *q, int rw, struct buffer_head *bh)
 {
 	int r;
+	struct dm_io *io;
 	struct mapped_device *md;
 
 	md = get_kdev(bh->b_rdev);
@@ -417,6 +638,7 @@ static int dm_request(request_queue_t *q, int rw, struct buffer_head *bh)
 		return 0;
 	}
 
+	io = alloc_io(md);
 	down_read(&md->lock);
 
 	r = __deferring(md, rw, bh);
@@ -425,7 +647,7 @@ static int dm_request(request_queue_t *q, int rw, struct buffer_head *bh)
 
 	else if (!r) {
 		/* not deferring */
-		r = __map_buffer(md, rw, bh);
+		r = __map_buffer(md, rw, bh, io);
 		if (r < 0)
 			goto bad;
 	} else
@@ -444,10 +666,12 @@ static int dm_request(request_queue_t *q, int rw, struct buffer_head *bh)
 
 static int check_dev_size(kdev_t dev, unsigned long block)
 {
+	unsigned int major = major(dev);
+	unsigned int minor = minor(dev);
+
 	/* FIXME: check this */
-	int minor = MINOR(dev);
-	unsigned long max_sector = (_block_size[minor] << 1) + 1;
-	unsigned long sector = (block + 1) * (_blksize_size[minor] >> 9);
+	unsigned long max_sector = (blk_size[major][minor] << 1) + 1;
+	unsigned long sector = (block + 1) * (blksize_size[major][minor] >> 9);
 
 	return (sector > max_sector) ? 0 : 1;
 }
@@ -460,6 +684,7 @@ static int __bmap(struct mapped_device *md, kdev_t dev, unsigned long block,
 {
 	struct buffer_head bh;
 	struct dm_target *ti;
+	union map_info map_context;
 	int r;
 
 	if (test_bit(DMF_BLOCK_IO, &md->flags)) {
@@ -470,18 +695,22 @@ static int __bmap(struct mapped_device *md, kdev_t dev, unsigned long block,
 		return -EINVAL;
 	}
 
+	if (!md->map)
+		return -EINVAL;
+
 	/* setup dummy bh */
 	memset(&bh, 0, sizeof(bh));
 	bh.b_blocknr = block;
 	bh.b_dev = bh.b_rdev = dev;
-	bh.b_size = _blksize_size[MINOR(dev)];
+	bh.b_size = blksize_size[major(dev)][minor(dev)];
 	bh.b_rsector = block * (bh.b_size >> 9);
 
 	/* find target */
 	ti = dm_table_find_target(md->map, bh.b_rsector);
 
 	/* do the mapping */
-	r = ti->type->map(ti, &bh, READ);
+	r = ti->type->map(ti, &bh, READ, &map_context);
+	ti->type->end_io(ti, &bh, READ, 0, &map_context);
 
 	if (!r) {
 		*r_dev = bh.b_rdev;
@@ -520,79 +749,19 @@ static int dm_user_bmap(struct inode *inode, struct lv_bmap *lvb)
 	return r;
 }
 
-/*-----------------------------------------------------------------
- * A bitset is used to keep track of allocated minor numbers.
- *---------------------------------------------------------------*/
-static spinlock_t _minor_lock = SPIN_LOCK_UNLOCKED;
-static struct mapped_device *_mds[MAX_DEVICES];
-
-static void free_minor(int minor)
+static void free_md(struct mapped_device *md)
 {
-	spin_lock(&_minor_lock);
-	_mds[minor] = NULL;
-	spin_unlock(&_minor_lock);
-}
-
-/*
- * See if the device with a specific minor # is free.
- */
-static int specific_minor(int minor, struct mapped_device *md)
-{
-	int r = -EBUSY;
-
-	if (minor >= MAX_DEVICES) {
-		DMWARN("request for a mapped_device beyond MAX_DEVICES (%d)",
-		       MAX_DEVICES);
-		return -EINVAL;
-	}
-
-	spin_lock(&_minor_lock);
-	if (!_mds[minor]) {
-		_mds[minor] = md;
-		r = minor;
-	}
-	spin_unlock(&_minor_lock);
-
-	return r;
-}
-
-static int next_free_minor(struct mapped_device *md)
-{
-	int i;
-
-	spin_lock(&_minor_lock);
-	for (i = 0; i < MAX_DEVICES; i++) {
-		if (!_mds[i]) {
-			_mds[i] = md;
-			break;
-		}
-	}
-	spin_unlock(&_minor_lock);
-
-	return (i < MAX_DEVICES) ? i : -EBUSY;
-}
-
-static struct mapped_device *get_kdev(kdev_t dev)
-{
-	struct mapped_device *md;
-
-	if (major(dev) != _major)
-		return NULL;
-
-	spin_lock(&_minor_lock);
-	md = _mds[minor(dev)];
-	if (md)
-		dm_get(md);
-	spin_unlock(&_minor_lock);
-
-	return md;
+	free_dev(md->dev);
+	mempool_destroy(md->io_pool);
+	kfree(md);
 }
 
 /*
  * Allocate and initialise a blank device with a given minor.
  */
-static struct mapped_device *alloc_dev(int minor)
+static struct mapped_device *alloc_md(kdev_t dev)
 {
+	int r;
 	struct mapped_device *md = kmalloc(sizeof(*md), GFP_KERNEL);
 
 	if (!md) {
@@ -600,29 +769,36 @@ static struct mapped_device *alloc_dev(int minor)
 		return NULL;
 	}
 
-	/* get a minor number for the dev */
-	minor = (minor < 0) ? next_free_minor(md) : specific_minor(minor, md);
-	if (minor < 0) {
+	memset(md, 0, sizeof(*md));
+
+	/* Allocate suitable device number */
+	if (!dev)
+		r = first_free_dev(md);
+	else
+		r = specific_dev(dev, md);
+
+	if (r) {
+		kfree(md);
+		return NULL;
+	}
+
+	md->io_pool = mempool_create(MIN_IOS, mempool_alloc_slab,
+				     mempool_free_slab, _io_cache);
+	if (!md->io_pool) {
+		free_md(md);
 		kfree(md);
 		return NULL;
 	}
 
-	memset(md, 0, sizeof(*md));
-	md->dev = mk_kdev(_major, minor);
 	init_rwsem(&md->lock);
 	atomic_set(&md->holders, 1);
 	atomic_set(&md->pending, 0);
 	init_waitqueue_head(&md->wait);
+	init_waitqueue_head(&md->eventq);
 
 	return md;
 }
 
-static void free_dev(struct mapped_device *md)
-{
-	free_minor(minor(md->dev));
-	kfree(md);
-}
-
 /*
  * The hardsect size for a mapped device is the largest hardsect size
  * from the devices it maps onto.
@@ -632,7 +808,7 @@ static int __find_hardsect_size(struct list_head *devices)
 	int result = 512, size;
 	struct list_head *tmp;
 
-	list_for_each(tmp, devices) {
+	list_for_each (tmp, devices) {
 		struct dm_dev *dd = list_entry(tmp, struct dm_dev, list);
 		size = get_hardsect_size(dd->dev);
 		if (size > result)
@@ -645,50 +821,63 @@ static int __find_hardsect_size(struct list_head *devices)
 /*
  * Bind a table to the device.
  */
+static void event_callback(void *context)
+{
+	struct mapped_device *md = (struct mapped_device *) context;
+
+	down_write(&md->lock);
+	md->event_nr++;
+	wake_up_interruptible(&md->eventq);
+	up_write(&md->lock);
+}
+
 static int __bind(struct mapped_device *md, struct dm_table *t)
 {
-	int minor = minor(md->dev);
+	unsigned int minor = minor(md->dev);
+	unsigned int major = major(md->dev);
 	md->map = t;
 
 	/* in k */
-	_block_size[minor] = dm_table_get_size(t) >> 1;
-	_blksize_size[minor] = BLOCK_SIZE;
-	_hardsect_size[minor] = __find_hardsect_size(dm_table_get_devices(t));
-	register_disk(NULL, md->dev, 1, &dm_blk_dops, _block_size[minor]);
+	blk_size[major][minor] = dm_table_get_size(t) >> 1;
+	blksize_size[major][minor] = BLOCK_SIZE;
+	hardsect_size[major][minor] =
+	    __find_hardsect_size(dm_table_get_devices(t));
+	register_disk(NULL, md->dev, 1, &dm_blk_dops, blk_size[major][minor]);
 
+	dm_table_event_callback(md->map, event_callback, md);
 	dm_table_get(t);
 	return 0;
 }
 
 static void __unbind(struct mapped_device *md)
 {
-	int minor = minor(md->dev);
+	unsigned int minor = minor(md->dev);
+	unsigned int major = major(md->dev);
 
-	dm_table_put(md->map);
-	md->map = NULL;
+	if (md->map) {
+		dm_table_event_callback(md->map, NULL, NULL);
+		dm_table_put(md->map);
+		md->map = NULL;
 
-	_block_size[minor] = 0;
-	_blksize_size[minor] = 0;
-	_hardsect_size[minor] = 0;
+	}
+
+	blk_size[major][minor] = 0;
+	blksize_size[major][minor] = 0;
+	hardsect_size[major][minor] = 0;
 }
 
 /*
  * Constructor for a new device.
  */
-int dm_create(int minor, struct dm_table *table, struct mapped_device **result)
+int dm_create(kdev_t dev, struct mapped_device **result)
 {
-	int r;
 	struct mapped_device *md;
 
-	md = alloc_dev(minor);
+	md = alloc_md(dev);
 	if (!md)
 		return -ENXIO;
 
-	r = __bind(md, table);
-	if (r) {
-		free_dev(md);
-		return r;
-	}
+	__unbind(md);	/* Ensure zero device size */
 
 	*result = md;
 	return 0;
@@ -702,8 +891,10 @@ void dm_get(struct mapped_device *md)
 void dm_put(struct mapped_device *md)
 {
 	if (atomic_dec_and_test(&md->holders)) {
+		if (md->map)
+			dm_table_suspend_targets(md->map);
 		__unbind(md);
-		free_dev(md);
+		free_md(md);
 	}
 }
 
@@ -731,8 +922,10 @@ int dm_swap_table(struct mapped_device *md, struct dm_table *table)
 
 	down_write(&md->lock);
 
-	/* device must be suspended */
-	if (!test_bit(DMF_SUSPENDED, &md->flags)) {
+	/*
+	 * The device must be suspended, or have no table bound yet.
+	 */
+	if (md->map && !test_bit(DMF_SUSPENDED, &md->flags)) {
 		up_write(&md->lock);
 		return -EPERM;
 	}
@@ -755,6 +948,7 @@ int dm_swap_table(struct mapped_device *md, struct dm_table *table)
  */
 int dm_suspend(struct mapped_device *md)
 {
+	int r = 0;
 	DECLARE_WAITQUEUE(wait, current);
 
 	down_write(&md->lock);
@@ -769,32 +963,39 @@ int dm_suspend(struct mapped_device *md)
 	}
 
 	set_bit(DMF_BLOCK_IO, &md->flags);
+	add_wait_queue(&md->wait, &wait);
 	up_write(&md->lock);
 
 	/*
 	 * Then we wait for the already mapped ios to
 	 * complete.
 	 */
-	down_read(&md->lock);
-
-	add_wait_queue(&md->wait, &wait);
+	run_task_queue(&tq_disk);
 	while (1) {
 		set_current_state(TASK_INTERRUPTIBLE);
 
-		if (!atomic_read(&md->pending))
+		if (!atomic_read(&md->pending) || signal_pending(current))
 			break;
 
 		schedule();
 	}
+	set_current_state(TASK_RUNNING);
 
-	current->state = TASK_RUNNING;
+	down_write(&md->lock);
 	remove_wait_queue(&md->wait, &wait);
-	up_read(&md->lock);
 
-	/* set_bit is atomic */
-	set_bit(DMF_SUSPENDED, &md->flags);
+	/* did we flush everything ? */
+	if (atomic_read(&md->pending)) {
+		clear_bit(DMF_BLOCK_IO, &md->flags);
+		r = -EINTR;
+	} else {
+		set_bit(DMF_SUSPENDED, &md->flags);
+		if (md->map)
+			dm_table_suspend_targets(md->map);
+	}
+	up_write(&md->lock);
 
-	return 0;
+	return r;
 }
 
 int dm_resume(struct mapped_device *md)
@@ -802,12 +1003,14 @@ int dm_resume(struct mapped_device *md)
 	struct deferred_io *def;
 
 	down_write(&md->lock);
-	if (!test_bit(DMF_SUSPENDED, &md->flags) ||
-	    !dm_table_get_size(md->map)) {
+	if (!test_bit(DMF_SUSPENDED, &md->flags)) {
 		up_write(&md->lock);
 		return -EINVAL;
 	}
 
+	if (md->map)
+		dm_table_resume_targets(md->map);
+
 	clear_bit(DMF_SUSPENDED, &md->flags);
 	clear_bit(DMF_BLOCK_IO, &md->flags);
 	def = md->deferred;
@@ -826,12 +1029,56 @@ struct dm_table *dm_get_table(struct mapped_device *md)
 
 	down_read(&md->lock);
 	t = md->map;
-	dm_table_get(t);
+	if (t)
+		dm_table_get(t);
 	up_read(&md->lock);
 
 	return t;
 }
 
+/*-----------------------------------------------------------------
+ * Event notification.
+ *---------------------------------------------------------------*/
+uint32_t dm_get_event_nr(struct mapped_device *md)
+{
+	uint32_t r;
+
+	down_read(&md->lock);
+	r = md->event_nr;
+	up_read(&md->lock);
+
+	return r;
+}
+
+int dm_add_wait_queue(struct mapped_device *md, wait_queue_t *wq,
+		      uint32_t event_nr)
+{
+	down_write(&md->lock);
+	if (event_nr != md->event_nr) {
+		up_write(&md->lock);
+		return 1;
+	}
+
+	add_wait_queue(&md->eventq, wq);
+	up_write(&md->lock);
+
+	return 0;
+}
+
+const char *dm_kdevname(kdev_t dev)
+{
+	static char buffer[32];
+	sprintf(buffer, "%03d:%03d", MAJOR(dev), MINOR(dev));
+	return buffer;
+}
+
+void dm_remove_wait_queue(struct mapped_device *md, wait_queue_t *wq)
+{
+	down_write(&md->lock);
+	remove_wait_queue(&md->eventq, wq);
+	up_write(&md->lock);
+}
+
 kdev_t dm_kdev(struct mapped_device *md)
 {
 	kdev_t dev;
@@ -861,8 +1108,8 @@ struct block_device_operations dm_blk_dops = {
 module_init(dm_init);
 module_exit(dm_exit);
 
-MODULE_PARM(major, "i");
-MODULE_PARM_DESC(major, "The major number of the device mapper");
 MODULE_DESCRIPTION(DM_NAME " driver");
 MODULE_AUTHOR("Joe Thornber <thornber@sistina.com>");
 MODULE_LICENSE("GPL");
+
+EXPORT_SYMBOL(dm_kdevname);
